import subprocess
from collections import defaultdict
from typing import Dict, List
from dagster.core.definitions.event_metadata import EventMetadataEntry

from dagster.core.events import DagsterEvent, DagsterEventType, EngineEventData
from dagster.core.executor.step_delegating import StepHandler
from dagster.core.executor.step_delegating.step_handler.base import StepHandlerContext
from dagster.serdes import serialize_dagster_namedtuple
from dagster.serdes.ipc import interrupt_ipc_subprocess_pid
import threading


class ProcessStepHandler(StepHandler):
    def __init__(self) -> None:
        super().__init__()
        self._step_pids: Dict[str, Dict[str, int]] = defaultdict(dict)
        self._step_pids_lock = threading.Lock()

    @property
    def name(self) -> str:
        return "ProcessStepHandler"

    def launch_step(self, step_handler_context: StepHandlerContext) -> List[DagsterEvent]:
        assert (
            len(step_handler_context.execute_step_args.step_keys_to_execute) == 1
        ), "Launching multiple steps is not currently supported"
        step_key = step_handler_context.execute_step_args.step_keys_to_execute[0]

        input_json = serialize_dagster_namedtuple(step_handler_context.execute_step_args)
        args = ["dagster", "api", "execute_step", input_json]
        p = subprocess.Popen(args)

        with self._step_pids_lock:
            self._step_pids[step_handler_context.execute_step_args.pipeline_run_id][
                step_key
            ] = p.pid
        return []

    def check_step_health(self, step_handler_context: StepHandlerContext) -> List[DagsterEvent]:
        # TODO not implemented
        return []

    def terminate_step(self, step_handler_context: StepHandlerContext) -> List[DagsterEvent]:
        assert (
            len(step_handler_context.execute_step_args.step_keys_to_execute) == 1
        ), "Launching multiple steps is not currently supported"
        step_key = step_handler_context.execute_step_args.step_keys_to_execute[0]

        with self._step_pids_lock:
            pid = self._step_pids[step_handler_context.execute_step_args.pipeline_run_id][step_key]

        events = [
            DagsterEvent(
                event_type_value=DagsterEventType.ENGINE_EVENT.value,
                pipeline_name=step_handler_context.execute_step_args.pipeline_origin.pipeline_name,
                step_key=step_key,
                message="Stopping process for step",
                event_specific_data=EngineEventData(
                    [
                        EventMetadataEntry.text(step_key, "Step key"),
                        EventMetadataEntry.int(pid, "Process id"),
                    ],
                ),
            )
        ]
        interrupt_ipc_subprocess_pid(pid)
        return events



from dagster.utils import merge_dicts
from gql.transport.requests import RequestsHTTPTransport

from ..headers.impl import get_dagster_cloud_api_headers
from ..storage.client import GqlShimClient


def create_cloud_dagit_client(url: str, api_token: str, retries=3):
    transport = RequestsHTTPTransport(
        url=f"{url}/graphql",
        use_json=True,
        headers=merge_dicts(
            {"Content-type": "application/json"},
            get_dagster_cloud_api_headers(api_token),
        ),
        retries=retries,
    )
    return GqlShimClient(transport=transport, fetch_schema_from_transport=True)



import click
import requests

from ...cli.utils import add_options, create_cloud_dagit_client_options
from ...errors import raise_http_error
from ...headers.impl import get_dagster_cloud_api_headers


@click.group(name="saml")
def saml_cli():
    """Customize your SAML settings."""


@saml_cli.command(name="upload-identity-provider-metadata")
@add_options(create_cloud_dagit_client_options)
@click.argument("metadata_file", type=click.File("r"), metavar="METADATA_FILE_PATH")
def upload_identity_provider_metadata_command(metadata_file, url: str, api_token: str):
    """Uploads your identity provider's metadata to enable SAML Single sign-on (SSO)."""

    response = requests.post(
        url=f"{url}/upload_idp_metadata",
        headers=get_dagster_cloud_api_headers(api_token),
        files={"metadata.xml": metadata_file},
    )

    raise_http_error(response)

    click.echo("The identity provider metadata was successfully uploaded.")



from contextlib import AbstractContextManager, ExitStack, contextmanager
from typing import Any, Dict

import requests
from dagster import Field, IntSource, Noneable, Permissive, StringSource
from dagster.utils import merge_dicts
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport
from requests.adapters import HTTPAdapter, Retry
from requests.exceptions import HTTPError

from ..errors import DagsterCloudHTTPError
from ..headers.impl import get_dagster_cloud_api_headers
from ..storage.errors import GraphQLStorageError


class GqlShimClient(AbstractContextManager):
    """Adapter for gql.Client that wraps errors in human-readable format."""

    def __init__(self, transport, fetch_schema_from_transport=True):
        self._exit_stack = ExitStack()

        try:
            # Use persistent connection when making requests
            self._client = self._exit_stack.enter_context(
                Client(transport=transport, fetch_schema_from_transport=fetch_schema_from_transport)
            )
        except HTTPError as e:
            raise DagsterCloudHTTPError(e)
        except Exception as exc:  # pylint: disable=broad-except
            raise GraphQLStorageError(exc.__str__()) from exc

    def execute(self, query: str, variable_values: Dict[str, Any] = None):
        try:
            return {"data": self._client.execute(gql(query), variable_values=variable_values)}
        except HTTPError as http_error:
            raise DagsterCloudHTTPError(http_error)
        except Exception as exc:  # pylint: disable=broad-except
            raise GraphQLStorageError(exc.__str__()) from exc

    def __enter__(self):
        return self

    def __exit__(self, _exception_type, _exception_value, _traceback):
        self._exit_stack.close()


def get_agent_headers(config_value: Dict[str, Any]):
    return get_dagster_cloud_api_headers(
        config_value["agent_token"],
        deployment_name=config_value.get("deployment"),
        additional_headers=config_value.get("headers"),
    )


@contextmanager
def create_agent_requests_session():
    with requests.Session() as session:
        adapter = HTTPAdapter(
            max_retries=Retry(total=8, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
        )
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        yield session


def create_proxy_client(url: str, config_value: Dict[str, Any], fetch_schema_from_transport=True):
    transport = RequestsHTTPTransport(
        url=url,
        use_json=True,
        headers=merge_dicts({"Content-type": "application/json"}, get_agent_headers(config_value)),
        verify=config_value.get("verify", True),
        retries=config_value.get("retries", 8),
        method=config_value.get("method", "POST"),
        timeout=config_value.get("timeout", None),
        cookies=config_value.get("cookies", {}),
    )
    return GqlShimClient(
        transport=transport, fetch_schema_from_transport=fetch_schema_from_transport
    )


def dagster_cloud_api_config():
    return {
        "url": Field(StringSource, is_required=True),
        "agent_token": Field(StringSource, is_required=True),
        "deployment": Field(StringSource, is_required=False),
        "headers": Field(Permissive(), default_value={}),
        "cookies": Field(Permissive(), default_value={}),
        "timeout": Field(Noneable(IntSource), default_value=None),
        "verify": Field(bool, default_value=True),
        # This may be how we want to implement our retry policy, but it may be too restrictive:
        # we may want to put this logic into the storage itself so that we can do some kind of
        # logging
        "retries": Field(IntSource, default_value=8),
        "method": Field(StringSource, default_value="POST"),
    }



import os

import click


def add_options(options):
    """
    Share options between commands
    https://stackoverflow.com/questions/40182157/shared-options-and-flags-between-commands
    """

    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


DEFAULT_DAGIT_URL = lambda: os.getenv(
    "DAGSTER_CLOUD_DAGIT_URL",
    "http://localhost:3000",
)


def _api_token():
    return os.getenv("DAGSTER_CLOUD_API_TOKEN")


def _show_api_token_prompt():
    password_exists = bool(_api_token())
    show_prompt = not password_exists
    return show_prompt


create_cloud_dagit_client_options = [
    click.option("--url", type=click.STRING, default=DEFAULT_DAGIT_URL, help="Dagit url"),
    click.option(
        "--api-token",
        type=click.STRING,
        default=_api_token,
        prompt=_show_api_token_prompt(),
        hide_input=True,
        required=True,
        help="API token generated in Dagit",
    ),
]



import os
import subprocess

from dagster.core.launcher.base import LaunchRunContext
from dagster.grpc.types import ExecuteRunArgs
from dagster.serdes.ipc import interrupt_ipc_subprocess_pid
from dagster.serdes.serdes import serialize_dagster_namedtuple

from . import WatchfulRunLauncher

PID_TAG = "process/pid"


class ProcessRunLauncher(WatchfulRunLauncher):
    def launch_run(self, context: LaunchRunContext) -> None:
        run = context.pipeline_run
        pipeline_code_origin = context.pipeline_code_origin

        input_json = serialize_dagster_namedtuple(
            ExecuteRunArgs(
                pipeline_origin=pipeline_code_origin,
                pipeline_run_id=run.run_id,
                instance_ref=self._instance.get_ref(),
            )
        )

        args = ["dagster", "api", "execute_run", input_json]
        process = subprocess.Popen(args)

        self._instance.add_run_tags(run.run_id, {PID_TAG: str(process.pid)})

    def _get_pid(self, run):
        if not run or run.is_finished:
            return None

        tags = run.tags

        if PID_TAG not in tags:
            return None

        return int(tags[PID_TAG])

    def can_terminate(self, run_id):
        run = self._instance.get_run_by_id(run_id)
        if not run:
            return False

        pid = self._get_pid(run)
        if not pid:
            return False

        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    def terminate(self, run_id):
        run = self._instance.get_run_by_id(run_id)
        if not run:
            return False

        pid = self._get_pid(run)
        if not pid:
            return False

        self._instance.report_run_canceling(run)

        interrupt_ipc_subprocess_pid(pid)

        return True

    def check_run_health(self, _run_id):
        pass



import click

from .saml.commands import saml_cli


def create_settings_cli_group():
    commands = {
        "saml": saml_cli,
    }

    @click.group(commands=commands)
    def group():
        """Customize your dagster-cloud settings."""

    return group


settings_cli = create_settings_cli_group()



import os
import sys
import threading
from collections import namedtuple

from dagster import check
from dagster.core.host_representation.grpc_server_registry import GrpcServerEndpoint
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster.daemon.daemon import get_default_daemon_logger
from dagster.grpc.client import DagsterGrpcClient, client_heartbeat_thread
from dagster.grpc.server import GrpcServerProcess
from dagster.serdes import ConfigurableClass, create_snapshot_id
from dagster_cloud.execution.step_handler.process_step_handler import ProcessStepHandler
from dagster_cloud.execution.watchful_run_launcher.process import ProcessRunLauncher

from .user_code_launcher import ReconcileUserCodeLauncher

CLEANUP_ZOMBIE_PROCESSES_INTERVAL = 5


class ProcessUserCodeEntry(
    namedtuple(
        "_ProcessUserCodeEntry",
        "grpc_server_process grpc_client heartbeat_shutdown_event heartbeat_thread",
    )
):
    def __new__(cls, grpc_server_process, grpc_client, heartbeat_shutdown_event, heartbeat_thread):
        return super(ProcessUserCodeEntry, cls).__new__(
            cls,
            check.inst_param(grpc_server_process, "grpc_server_process", GrpcServerProcess),
            check.inst_param(grpc_client, "grpc_client", DagsterGrpcClient),
            check.inst_param(heartbeat_shutdown_event, "heartbeat_shutdown_event", threading.Event),
            check.inst_param(heartbeat_thread, "heartbeat_thread", threading.Thread),
        )


class ProcessUserCodeLauncher(ReconcileUserCodeLauncher, ConfigurableClass):
    def __init__(self, inst_data=None, wait_for_processes=False):
        self._inst_data = inst_data
        self._logger = get_default_daemon_logger("ProcessDagsterUserCodeLauncher")

        # Dict[int, ProcessUserCodeEntry], map from pid to all servers being spun up (including
        # old servers in the process of being shut down)
        self._process_entries = {}

        # Dict[str, int], map from hash of location_name + metadata (see _get_process_key)
        # to the pid for that location-metadata combination
        self._active_pids = {}

        self._heartbeat_ttl = 60
        self._wait_for_processes = wait_for_processes

        self._cleanup_zombies_shutdown_event = threading.Event()
        self._cleanup_zombies_thread = None

        self._step_handler = (
            ProcessStepHandler()
        )  # the process handler keeps pids in memory, so we keep a single instance of it

        super(ProcessUserCodeLauncher, self).__init__()

    def start(self):
        super().start()
        self._cleanup_zombies_thread = threading.Thread(
            target=self._cleanup_zombie_processes,
            args=(self._cleanup_zombies_shutdown_event,),
            name="cleanup-zombie-processes",
            daemon=True,
        )
        self._cleanup_zombies_thread.start()

    def _cleanup_zombie_processes(self, shutdown_event):
        while True:
            shutdown_event.wait(CLEANUP_ZOMBIE_PROCESSES_INTERVAL)
            if shutdown_event.is_set():
                break

            # Clean up any child processes that have finished since last check
            while True:
                # This may need to be different on Windows because process groups are
                # handled differently.
                try:
                    pid, _exit_code = os.waitpid(0, os.WNOHANG)
                except ChildProcessError:
                    # Raised when there are no child processes
                    break

                if pid == 0:
                    break

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @staticmethod
    def from_config_value(inst_data, config_value):
        return ProcessUserCodeLauncher(inst_data=inst_data)

    def _add_server(self, location_name, metadata):
        process_key = self._get_process_key(location_name, metadata)
        existing_pid = self._get_existing_pid(location_name, metadata)
        if existing_pid:
            self._remove_pid(existing_pid)
            del self._active_pids[process_key]

        return self._start_new_server(location_name, metadata)

    def _start_new_server(self, location_name, metadata):
        process_key = self._get_process_key(location_name, metadata)
        loadable_target_origin = self._get_loadable_target_origin(metadata)
        server_process = GrpcServerProcess(
            loadable_target_origin=loadable_target_origin,
            heartbeat=True,
            heartbeat_timeout=self._heartbeat_ttl,
        )
        server_id = self._wait_for_server(
            host="localhost", port=server_process.port, socket=server_process.socket
        )

        client = DagsterGrpcClient(
            port=server_process.port,
            socket=server_process.socket,
            host="localhost",
            use_ssl=False,
        )

        heartbeat_shutdown_event = threading.Event()
        heartbeat_thread = threading.Thread(
            target=client_heartbeat_thread,
            args=(client, heartbeat_shutdown_event),
        )
        heartbeat_thread.daemon = True
        heartbeat_thread.start()

        pid = server_process.pid

        self._process_entries[server_process.pid] = ProcessUserCodeEntry(
            server_process,
            client,
            heartbeat_shutdown_event,
            heartbeat_thread,
        )
        self._active_pids[process_key] = pid

        endpoint = GrpcServerEndpoint(
            server_id=server_id,
            host="localhost",
            port=server_process.port,
            socket=server_process.socket,
        )

        return endpoint

    def _get_process_key(self, location_name, metadata):
        return f"{location_name}_{create_snapshot_id(metadata)[0:6]}"

    def _get_loadable_target_origin(self, metadata):
        return LoadableTargetOrigin(
            executable_path=sys.executable,
            python_file=metadata.python_file,
            package_name=metadata.package_name,
        )

    def _get_existing_pid(self, location_name, metadata):
        process_key = self._get_process_key(location_name, metadata)
        return self._active_pids.get(process_key)

    def _gen_update_server(self, location_name, old_metadata, new_metadata):
        old_process_key = self._get_process_key(location_name, old_metadata)
        new_process_key = self._get_process_key(location_name, new_metadata)
        existing_pid = self._get_existing_pid(location_name, old_metadata)

        updated_server = self._start_new_server(location_name, new_metadata)
        yield updated_server

        if existing_pid:
            self._logger.info(
                "Stopping old process for location {location_name}".format(
                    location_name=location_name,
                )
            )
            self._remove_pid(existing_pid)
            if old_process_key != new_process_key:
                del self._active_pids[old_process_key]

    def _remove_server(self, location_name, metadata):
        process_key = self._get_process_key(location_name, metadata)
        existing_pid = self._get_existing_pid(location_name, metadata)
        if existing_pid:
            self._logger.info(
                "Stopping process for location {location_name}".format(
                    location_name=location_name,
                )
            )
            self._remove_pid(existing_pid)
            del self._active_pids[process_key]

    def _remove_pid(self, pid):
        process_entry = self._process_entries[pid]
        process_entry.heartbeat_shutdown_event.set()
        process_entry.heartbeat_thread.join()
        # Rely on heartbeat failure to eventually kill the process
        del self._process_entries[pid]

    def step_handler(self):
        return self._step_handler

    def run_launcher(self):
        launcher = ProcessRunLauncher()
        launcher.register_instance(self._instance)

        return launcher

    def __exit__(self, exception_type, exception_value, traceback):
        super().__exit__(exception_value, exception_value, traceback)

        while len(self._process_entries):
            pid = next(iter(self._process_entries))
            process_entry = self._process_entries[pid]

            self._remove_pid(pid)
            if self._wait_for_processes:
                process_entry.grpc_client.shutdown_server()
                process_entry.grpc_server_process.wait()

        if self._cleanup_zombies_thread:
            self._cleanup_zombies_shutdown_event.set()
            self._cleanup_zombies_thread.join()
